// Written 2000 by Michael Pfeiffer
#include "GhostscriptWrapper.h"

#include <stdio.h>

#include <be/interface/Window.h>
#include <be/app/Messenger.h>
#include <be/kernel/image.h>

extern "C" {
#include "gsdll.h"
}

// Set macro MORE_DEBUG to 1 to enable debugging output
#define MORE_DEBUG 0

// Macros
#if MORE_DEBUG
	#define OUT(args) fprintf args
#else
	#define OUT(args)
#endif

#define CHECK_GS_DLL if (!gGhostscriptLoaded) return

// Constants
const char* gExitLoop = "***exit_loop***";

// Global Variables
int ( * gGsdllRevision)     (char  **, char **, long *, long *);
int ( * gGsdllInit)         (GSDLL_CALLBACK, HWND, int argc, char ** argv);
int ( * gGsdllExecuteBegin) (void);
int ( * gGsdllExecuteCont)  (const char * str, int len);
int ( * gGsdllExecuteEnd)   (void);
int ( * gGsdllExit)         (void);
int ( * gGsdllLockDevice)   (unsigned char *, int);
void (* gGsdllDraw)         (unsigned char *device, BView *view, BRect dest, BRect src);


bool GetImageSymbol(image_id image, const char* name, void** function)
{
	if (get_image_symbol(image, name, B_SYMBOL_TYPE_TEXT, function) != B_OK) {
		fprintf(stderr, "could not get image symbol %s \n", name);
		return false;
	}
	return true;
}

bool Ghostscript::LoadGSDll() {
	if (gGhostscriptLoaded) return true;
	image_id gs;
	gs = load_add_on("/boot/home/config/lib/libgs.so");
	if (gs > 0) {
		if (GetImageSymbol(gs, "gsdll_revision",      (void**)&gGsdllRevision) &&
			GetImageSymbol(gs, "gsdll_init",          (void**)&gGsdllInit) &&
			GetImageSymbol(gs, "gsdll_execute_begin", (void**)&gGsdllExecuteBegin) &&
			GetImageSymbol(gs, "gsdll_execute_cont",  (void**)&gGsdllExecuteCont) &&
			GetImageSymbol(gs, "gsdll_execute_end",   (void**)&gGsdllExecuteEnd) &&
			GetImageSymbol(gs, "gsdll_exit",          (void**)&gGsdllExit) &&
			GetImageSymbol(gs, "gsdll_lock_device",   (void**)&gGsdllLockDevice) &&
			GetImageSymbol(gs, "gsdll_draw",          (void**)&gGsdllDraw)) 
		{
			gGhostscriptLoaded = true;
			return true;
		}
	}
	return false;
}



bool Ghostscript::gGhostscriptLoaded = false;
unsigned char *Ghostscript::fBeDevice = NULL;
sem_id Ghostscript::fWaitForPageSem = -1;
BLooper *Ghostscript::looper = NULL;
bool Ghostscript::fStop = false;

void Ghostscript::SendMessage(uint32 what) {
	CHECK_GS_DLL;
	if (looper) {
		BMessenger m(NULL, looper);
		m.SendMessage(what);
	}
}

void Ghostscript::SendMessage(BMessage *msg) {
	CHECK_GS_DLL;
	if (looper) {
		BMessenger m(NULL, looper);
		m.SendMessage(msg);
	}
}

void Ghostscript::Init() {
	CHECK_GS_DLL;
	fIsRunning = false;
	fPaperWidth = 800; fPaperHeight = 600; fXDpi = 72; fYDpi = 72;
	fDeviceName = "beos";
}

Ghostscript::Ghostscript() {
	LoadGSDll();
	Init();
}

Ghostscript::~Ghostscript() {
	Stop();
}

int32 ghostscript_thread(void *data) {
	Ghostscript *gs = (Ghostscript*)data;
	gs->Run();
	return 0;
}

void Ghostscript::Start() {
	CHECK_GS_DLL;
	if (!fIsRunning) {		
		fCmdSendSem = create_sem(1, "cmd_send_sem");
		fCmdReceiveSem = create_sem(0, "cmd_receive_sem");
		fWaitForPageSem = create_sem(0, "wait_for_page_sem");
		fIsRunning = true;
		fStop = false;
		fThread = spawn_thread(ghostscript_thread, "ghostscript_thread", B_NORMAL_PRIORITY, this);
		resume_thread(fThread);
	}
}

void Ghostscript::Stop() {
	CHECK_GS_DLL;
	if (fIsRunning) {
		status_t exit_status;
		Send(gExitLoop, strlen(gExitLoop));
		wait_for_thread(fThread, &exit_status);
	}
}

void Ghostscript::Run() {
	CHECK_GS_DLL;
	int myargc = 8;
	char *myargv[] = {
		"gs",
		NULL, // "-dDEVICEWIDTH=300",
		NULL, // "-dDEVICEHEIGHT=200",
		NULL, // "-sDEVICE=beos",
		NULL, // "-dDEVICEXRESOLUTION=72",
		NULL, // "-dDEVICEYRESOLUTION=72",
		"-dNOPAUSE",
//		"-dSAFER",
//		"-dBitsPerPixel=32",
		"-q",
		NULL
	};

	char buffer[250];
	myargv[1] = buffer +   0; sprintf(buffer +   0, "-dDEVICEWIDTH=%d", (int)fPaperWidth); 
	myargv[2] = buffer +  50; sprintf(buffer +  50, "-dDEVICEHEIGHT=%d", (int)fPaperHeight); 
//	myargv[1] = buffer +   0; sprintf(buffer +   0, "-dDEVICEWIDTH=%d", fBitmapWidth); 
//	myargv[2] = buffer +  50; sprintf(buffer +  50, "-dDEVICEHEIGHT=%d", fBitmapHeight); 
	myargv[3] = buffer + 100; sprintf(buffer + 100, "-sDEVICE=%s", fDeviceName.String()); 
	myargv[4] = buffer + 150; sprintf(buffer + 150, "-dDEVICEXRESOLUTION=%d", (int)fXDpi); 
	myargv[5] = buffer + 200; sprintf(buffer + 200, "-dDEVICEYRESOLUTION=%d", (int)fYDpi); 

	int32 code = gGsdllInit(gsdll_callback, NULL, myargc, myargv);
	OUT((stderr, "gsdll_init returns %d\n", (int)code));
	if (code == 0) {
		code = gGsdllExecuteBegin();
		OUT((stderr, "gsdll_execute_begin returns %d\n", (int)code));
	}
	while (code == 0) {
		acquire_sem(fCmdReceiveSem);
		if (fCommand != gExitLoop) {
			code = gGsdllExecuteCont(fCommand.String(), fCommand.Length());
			SendMessage(((code == 0) || fStop) ? CMD_NOTIFY : CMD_ERROR_NOTIFY);
			OUT((stderr, "gsdll_execute_cont returns %d\n", (int)code));
			release_sem(fCmdSendSem);
		} else {
			break;
		}
	}
	fIsRunning = false;
	delete_sem(fCmdSendSem);
	delete_sem(fCmdReceiveSem);
	delete_sem(fWaitForPageSem);

	if (code > -100) {
		code = gGsdllExecuteEnd();
	 	OUT((stderr, "gsdll_execute_end returns %d\n", (int)code));
	}
	code = gGsdllExit();
	OUT((stderr, "gsdll_exit returns %d\n", (int)code));
	SendMessage(EXIT_NOTIFY);
}

bool Ghostscript::Send(const char *cmd, int32 len) {
	if (fIsRunning && (B_OK == acquire_sem(fCmdSendSem))) {
		this->fCommand.SetTo(cmd, len+1);
		release_sem(fCmdReceiveSem);
	}
	return false;
}

bool Ghostscript::Draw(BView *view, BRect from, BRect to) {
	CHECK_GS_DLL false;
	if (fIsRunning) {
		if (fBeDevice) {
			gGsdllLockDevice(fBeDevice, 1);
			gGsdllDraw(fBeDevice, view, to, from);
			view->Sync();
			gGsdllLockDevice(fBeDevice, 0);
			return true;
		}
	}
	return false;
}

int Ghostscript::gsdll_callback(int message, char *str, unsigned long count) {
	switch (message) {
	case GSDLL_STDIN:
		OUT((stderr, "GSDLL_STDIN called\n"));
		return 0; 
	case GSDLL_STDOUT:
		OUT((stderr, "GSDLL_STDOUT called\n"));
		if (str != (char *)NULL) {
			fwrite(str, 1, count, stderr);
		}
		return count;
	case GSDLL_DEVICE:
		OUT((stderr, "GSDLL_DEVICE called\n"));
		if (count) {
			fBeDevice = (unsigned char *)str;
		} else {
			fBeDevice = NULL;
		}
		OUT((stderr, "BDEV=%p\n", str));
		break;
    case GSDLL_SYNC:
        OUT((stderr,"Callback: SYNC %p\n", str));
		Ghostscript::SendMessage(Ghostscript::SYNC_NOTIFY);
		break;
	case GSDLL_PAGE: 
		OUT((stderr,"Callback: PAGE %p\n", str));
		Ghostscript::SendMessage(Ghostscript::PAGE_NOTIFY);
		acquire_sem(fWaitForPageSem);
		if (fStop) {
			return -1;
		} else { 
			Ghostscript::SendMessage(Ghostscript::PAGE_CONT_NOTIFY);
		}
		break;
	case GSDLL_SIZE: {
		OUT((stderr,"Callback: SIZE %p width=%d height=%d\n", str,
			(int)(count & 0xffff), (int)((count>>16) & 0xffff)) );
		BMessage msg(SIZE_NOTIFY);
		msg.AddInt32("width", (int)(count & 0xffff));
		msg.AddInt32("height", (int)((count>>16) & 0xffff));
		Ghostscript::SendMessage(&msg);
		}
		break;
	case GSDLL_POLL:
		OUT((stderr,"Callback: POLL\n"));
		return 0; /* no error */
	default:
		OUT((stderr,"Callback: Unknown message=%d\n",message));
		break;
	}
	return 0;
}

void Ghostscript::CalculateBitmapSize() {
	fBitmapWidth = fXDpi * fPaperWidth / 72;
	fBitmapHeight = fYDpi * fPaperHeight / 72;
}

void Ghostscript::SetPaperSize(int32 width, int32 height) {
	fPaperWidth = width; fPaperHeight = height;
	CalculateBitmapSize();
}

void Ghostscript::SetDpi(int32 xdpi, int32 ydpi) {
	fXDpi = xdpi; fYDpi = ydpi;
	CalculateBitmapSize();
}

void Ghostscript::SetDevice(const char *dev) {
	fDeviceName = dev;
}

bool Ghostscript::IsProcessing() {
	if (fIsRunning) {
		int32 count;
		if (B_NO_ERROR == get_sem_count(fCmdSendSem, &count)) {
			return count <= 0;
		}
	}
	return false;
}

void Ghostscript::Wait() {
	if (IsProcessing()) {
		acquire_sem(fCmdSendSem);
		release_sem(fCmdSendSem);
	}
}

bool Ghostscript::WaitsForPage() {
	if (fIsRunning) {
		int32 count;
		if (B_NO_ERROR == get_sem_count(fWaitForPageSem, &count)) {
			return count <= -1;
		}
	}
	return false;
}

void Ghostscript::NextPage(bool stop) {
	if (WaitsForPage()) {
		fStop = stop;
		release_sem(fWaitForPageSem);
	}
}
